"""Code for managing and configuring a dbt project."""
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import List, Optional, Union


class DbtResourceType(Enum):
    """
    Enum class that represents a dbt resource type.
    """

    MODEL = auto()
    SEED = auto()
    SNAPSHOT = auto()


@dataclass
class DbtProject:
    """
    Class that represents a dbt project configuration.

    Args:
        name: Name of dbt project,
        project_dir: Path to directory that contains dbt project.yml
        profiles_dir: Path to directory that contains dbt profiles.yml
    """

    name: str
    project_dir: Union[str, Path]
    profiles_dir: Union[str, Path]


@dataclass
class DbtProfile:
    """
    Class that represents a dbt profile configuration.

    Args:
        target: dbt target, usualy "dev" or "prod"
    """

    target: str


@dataclass
class DbtNode:
    """
    Class that represents a dbt node in the project.

    Args:
        name: dbt node name, e. my_model_a
        unique_id: dbt id e. model.sample_project.my_model_a
        resource_type: dbt resource type, e. model or seed
        depends_on: e. ["model.sample_project.my_model_b"]
        has_tests: if node is a test
    """

    name: str
    unique_id: str
    resource_type: DbtResourceType
    depends_on: List[str]
    has_tests: bool = False


@dataclass
class DbtDagOptions:
    """
    Class to add dbt DAG configurations.

    Args:
        select: dbt module to include in the run
        exclude: dbt module to exclude in the run
        run_test_after_model: run test afeter run model
    """

    select: Optional[str] = None
    exclude: Optional[str] = None
    run_test_after_model: bool = False



"""Utility functions for interacting with dbt using command-line commands."""
import shutil
from typing import Optional

from prefect_dbt_flow.dbt import DbtDagOptions, DbtProfile, DbtProject
from prefect_dbt_flow.utils import cmd

DBT_EXE = shutil.which("dbt") or "dbt"


def dbt_ls(
    project: DbtProject,
    dag_options: Optional[DbtDagOptions],
    output: str = "json",
) -> str:
    """
    Code that lists resources from the dbt project, using `dbt ls` command.

    Args:
        project: A class that represents a dbt project configuration.
        dag_options: A class to add dbt DAG configurations.
        output: Format of output, default is JSON.

    Returns:
        list of JSON objects containing dbt resources.
    """
    dbt_ls_cmd = [DBT_EXE, "ls"]
    dbt_ls_cmd.extend(["--project-dir", str(project.project_dir)])
    dbt_ls_cmd.extend(["--profiles-dir", str(project.profiles_dir)])
    dbt_ls_cmd.extend(["--output", output])

    if dag_options:
        if dag_options.select:
            dbt_ls_cmd.extend(["--select", dag_options.select])
        if dag_options.exclude:
            dbt_ls_cmd.extend(["--exclude", dag_options.exclude])

    return cmd.run(" ".join(dbt_ls_cmd))


def dbt_run(project: DbtProject, profile: DbtProfile, model: str) -> str:
    """
    Function that executes `dbt run` command

    Args:
        project: A class that represents a dbt project configuration.
        profile: A class that represents a dbt profile configuration.
        model: Name of the model to run.

    Returns:
        A string representing the output of the `dbt run` command.
    """
    dbt_run_cmd = [DBT_EXE, "run"]
    dbt_run_cmd.extend(["--project-dir", str(project.project_dir)])
    dbt_run_cmd.extend(["--profiles-dir", str(project.profiles_dir)])
    dbt_run_cmd.extend(["-t", profile.target])
    dbt_run_cmd.extend(["-m", model])

    return cmd.run(" ".join(dbt_run_cmd))


def dbt_test(project: DbtProject, profile: DbtProfile, model: str) -> str:
    """
    Function that executes `dbt test` command

    Args:
        project: A class that represents a dbt project configuration.
        profile: A class that represents a dbt profile configuration.
        model: Name of the model to run.

    Returns:
        A string representing the output of the `dbt test` command.
    """
    dbt_test_cmd = [DBT_EXE, "test"]
    dbt_test_cmd.extend(["--project-dir", str(project.project_dir)])
    dbt_test_cmd.extend(["--profiles-dir", str(project.profiles_dir)])
    dbt_test_cmd.extend(["-t", profile.target])
    dbt_test_cmd.extend(["-m", model])

    return cmd.run(" ".join(dbt_test_cmd))


def dbt_seed(project: DbtProject, profile: DbtProfile, seed: str) -> str:
    """
    Function that executes `dbt seed` command

    Args:
        seed: Name of the seed to run.

    Returns:
        A string representing the output of the `dbt seed` command
    """
    dbt_seed_cmd = [DBT_EXE, "seed"]
    dbt_seed_cmd.extend(["--project-dir", str(project.project_dir)])
    dbt_seed_cmd.extend(["--profiles-dir", str(project.profiles_dir)])
    dbt_seed_cmd.extend(["-t", profile.target])
    dbt_seed_cmd.extend(["--select", seed])

    return cmd.run(" ".join(dbt_seed_cmd))



"""Code for parsing dbt project and generate a list of dbt nodes"""
import json
from typing import List, Optional

from prefect_dbt_flow.dbt import (
    DbtDagOptions,
    DbtNode,
    DbtProject,
    DbtResourceType,
    cli,
)


def parse_dbt_project(
    project: DbtProject, dag_options: Optional[DbtDagOptions] = None
) -> List[DbtNode]:
    """
    Parses a list of dbt nodes class objects from dbt ls cli command.

    Args:
        project: A class that represents a dbt project configuration.
        dag_options: A class to add dbt DAG configurations.

    Returns:
        dbt_graph: A list of dbt nodes, each node as a dataclass.
    """
    dbt_graph: List[DbtNode] = []
    models_with_tests: List[str] = []

    dbt_ls_output = cli.dbt_ls(project, dag_options)

    for line in dbt_ls_output.split("\n"):
        try:
            node_dict = json.loads(line.strip())

            if node_dict["resource_type"] == "model":
                dbt_graph.append(
                    DbtNode(
                        name=node_dict["name"],
                        unique_id=node_dict["unique_id"],
                        resource_type=DbtResourceType.MODEL,
                        depends_on=node_dict["depends_on"].get("nodes", []),
                    )
                )
            if node_dict["resource_type"] == "test":
                models_with_tests.extend(node_dict["depends_on"]["nodes"])

            if node_dict["resource_type"] == "seed":
                dbt_graph.append(
                    DbtNode(
                        name=node_dict["name"],
                        unique_id=node_dict["unique_id"],
                        resource_type=DbtResourceType.SEED,
                        depends_on=node_dict["depends_on"].get("nodes", []),
                    )
                )

        except json.decoder.JSONDecodeError:
            pass

    # Check if a node has tests
    for dbt_node in dbt_graph:
        if dbt_node.unique_id in models_with_tests:
            dbt_node.has_tests = True

    # Remove dependencies if not in Graph (needed in case of select/exclude)
    all_model_ids = [dbt_node.unique_id for dbt_node in dbt_graph]
    for dbt_node in dbt_graph:
        dbt_node.depends_on = [
            node_id for node_id in dbt_node.depends_on if node_id in all_model_ids
        ]

    return dbt_graph



"""Prefect dbt flow initialization"""
# isort: skip_file
# ruff: noqa: F401
# fmt: off
from prefect_dbt_flow.dbt import (
    DbtDagOptions,
    DbtNode,
    DbtProfile,
    DbtProject
    )
from prefect_dbt_flow.flow import dbt_flow
# fmt: on



"""Functions to create a prefect flow for a dbt project."""
from typing import Optional

from prefect import Flow, flow

from prefect_dbt_flow.dbt import DbtDagOptions, DbtProfile, DbtProject, graph, tasks


def dbt_flow(
    project: DbtProject,
    profile: Optional[DbtProfile] = None,
    dag_options: Optional[DbtDagOptions] = None,
    flow_kwargs: Optional[dict] = None,
) -> Flow:
    """
    Create a PrefectFlow for executing a dbt project.

    Args:
        project: A Class that represents a dbt project configuration.
        profile: A Class that represents a dbt profile configuration.
        dag_options: A Class to add dbt DAG configurations.
        flow_kwargs: A dict of prefect @flow arguments

    Returns:
        dbt_flow: A Prefec Flow.
    """
    all_flow_kwargs = {
        "name": project.name,
        **(flow_kwargs or {}),
    }

    dbt_graph = graph.parse_dbt_project(project, dag_options)

    @flow(**all_flow_kwargs)
    def dbt_flow():
        """
        Function that configurates and runs a Prefect flow.

        Returns:
            A prefect flow
        """
        tasks.generate_tasks_dag(
            project,
            profile,
            dbt_graph,
            dag_options.run_test_after_model if dag_options else False,
        )

    return dbt_flow
